#include "crossroads.h"

void servesockets (int clientsock, int backendsock) {
    fd_set readset, exceptset;
    int pid, src_fd, dst_fd,
	nread, totread = 0,
	nwritten, totwritten,
	i, nfd;
    char buf[2048], *src_str, *dst_str;
    time_t start_t;
    struct timeval tv, *tvp;

    if (! flag_foreground) {
	if ( (pid = fork()) < 0 )
	    /* Fork failure */
	    error ("Failed to fork: %s", strerror(errno));
	else if (pid) {
	    /* Parent branch */
	    msg ("Socket communications handled by pid %d", pid);
	    return;
	} else {
	    /* Child branch: here we will serve the socket pair.
	     * We want to catch interrupts, but not die. */
	    for (i = 0; relevant_sigs[i]; i++)
		signal (relevant_sigs[i], interrupt);

	    /* Forked for the second time... */
	    program_stage = stage_serving;

	    set_program_title ("crossroads: servicing %s to %s",
			       activeservice->name,
			       activeservice->backend[current_backend].name);
	}
    }

    /* Promote the verbosity of the backend level, log time, etc.. */
    flag_verbose = activeservice->backend[current_backend].verbosity;
    start_t = time(0);
    lock_reporter();
    servicereport->nclients++;
    servicereport->backendstate[current_backend].nclients++;
    unlock_reporter();

    /* Cpio source to dest and back again. */
    while (1) {

	/* Prepare file descriptor sets for select(3). */
	FD_ZERO (&readset);
	FD_SET (clientsock, &readset);
	FD_SET (backendsock, &readset);

	FD_ZERO (&exceptset);
	FD_SET (clientsock, &exceptset);
	FD_SET (backendsock, &exceptset);

	/* Wait for a readable fd (or for an exception...)
	 * optionally using the specified timeout.
	 */
	if (activeservice->sessiontimeout) {
	    tv.tv_sec  = activeservice->sessiontimeout;
	    tv.tv_usec = 0;
	    tvp = &tv;
	} else
	    tvp = 0;

	nfd = select (FD_SETSIZE, &readset, 0, &exceptset, tvp);
	if (nfd < 1) {
	    /* There isn't even one FD for a session. Decrease the count. */
	    lock_reporter();
	    servicereport->nclients--;
	    servicereport->backendstate[current_backend].nclients--;
	    unlock_reporter();
	    
	    if (! nfd) {
		/* Truly a session end. We're all done. This _must_
		 * be due to a timeout, or the read() below would have
		 * signalled EOF. */
		warning ("Timout in service %s (backend %s) after %d secs",
			 activeservice->name,
			 activeservice->backend[current_backend].name,
			 activeservice->sessiontimeout);
		mark_activity (totread, time(0) - start_t, st_available);
		exit (0);
	    } else {
		/* nfd < 0: This is an error condition.
		 * select() has failed! */
		error ("Failed to wait for network input: %s",
		       strerror(errno));
	    }
	}

	/* Check for exceptions. A backend-side exception will be reported
	 * as a failure.
	 */
	if (FD_ISSET(clientsock, &exceptset)) {
	    lock_reporter();
	    servicereport->nclients--;
	    servicereport->backendstate[current_backend].nclients--;
	    unlock_reporter();
	    error ("Exception on client connection");
	}
	if (FD_ISSET(backendsock, &exceptset)) {
	    lock_reporter();
	    servicereport->nclients--;
	    servicereport->backendstate[current_backend].nclients--;
	    unlock_reporter();
	    mark_activity (totread, time(0) - start_t, st_unavailable);
	    error ("Exception on backend connection");
	}

	/* Determine which socket is readable. We'll write to the other one. */
	if (FD_ISSET(clientsock, &readset)) {
	    src_fd = clientsock;
	    dst_fd = backendsock;
	    src_str = "client";
	    dst_str = "backend";
	} else if (FD_ISSET(backendsock, &readset)) {
	    src_fd = backendsock;
	    dst_fd = clientsock;
	    src_str = "backend";
	    dst_str = "client";
	}

	/* Read source socket. */
	if (! (nread = read (src_fd, buf, sizeof(buf))) ) {
	    lock_reporter();
	    servicereport->nclients--;
	    servicereport->backendstate[current_backend].nclients--;
	    unlock_reporter();	    
	    mark_activity (totread, time(0) - start_t, st_available);
	    close (backendsock);
	    close (clientsock);

	    /* In forked mode, we're done. In foreground mode, RTC. */
	    if (flag_foreground)
		return;
	    else
		exit (0);
	} else if (nread < 0) {
	    if (src_fd == backendsock)
		mark_activity (totread, time(0) - start_t, st_unavailable);
	    lock_reporter();
	    servicereport->nclients--;
	    servicereport->backendstate[current_backend].nclients--;
	    unlock_reporter();
	    error ("Read error on connection from %s", src_str);
	}
	
	totread += nread;

	msg ("Read %d bytes from %s", nread, src_str);

	trafficlog (buf, nread, src_fd == clientsock);
	thruputlog (buf, nread, src_fd == clientsock);

	/* Write in chunks to dest socket. */
	totwritten = 0;
	while (totwritten < nread) {
	    if ( (nwritten = write (dst_fd,
				    buf + totwritten,
				    nread - totwritten)) < 1 ) {
		if (dst_fd == backendsock)
		    mark_activity (totread, time(0) - start_t, st_unavailable);
		lock_reporter();
		servicereport->nclients--;
		servicereport->backendstate[current_backend].nclients--;
		unlock_reporter();
		error ("Write error when copying to %s", dst_str);
	    }
	    msg ("Wrote %d bytes to %s", nwritten, dst_str);
	    totwritten += nwritten;
	}

	msg ("Copied total of %d bytes from %s to %s",
	     nread, src_str, dst_str);
    }
}
